/*
    Musagi: a converter from XM to Panasonic Music

    18.3.2012: First somewhat functional version

    xm2moddi:
    30.8.2004 Marq
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define U8() fgetc(f)
#define U16() (fgetc(f)+(fgetc(f)<<8))
#define U32() (fgetc(f)+(fgetc(f)<<8)+(fgetc(f)<<16)+(fgetc(f)<<24))

#define M8(x) fputc(x,f)
#define M16(x) fputc(x>>8,f);fputc(x&0xff,f)
#define M32(x) fputc((x>>24)&255,f);fputc((x>>16)&255,f);fputc((x>>8)&255,f);\
fputc(x&255,f)

#define charset(x) ((x>='!') && (x<='z'))

typedef struct ptn_t
{
    unsigned short rows,packed;
    unsigned char *pdata;
    unsigned char *data;
} PTN;

typedef struct sample_t
{
    unsigned len,vol,loop,looplen;
    char *name;
    int finetune;
    signed char *data;
} SMP;

void convptn(PTN *ptn,int chn)
{
    unsigned char *s,*d;
    int n,i,j;

    ptn->data=malloc(chn*5*ptn->rows);
    s=ptn->pdata;
    d=ptn->data;

    memset(d,0,chn*5*ptn->rows);

    if(!ptn->packed)
        return;

    for(n=0;n<ptn->rows;n++)
    {
      for(j=0;j<chn;j++)
      {
        unsigned char b;
        b=*s;
        if(b&0x80) // Packed
        {
            s++;
            if(b&1) *d++=*s++; else *d++;
            if(b&2) *d++=*s++; else *d++;
            if(b&4) *d++=*s++; else *d++;
            if(b&8) *d++=*s++; else *d++;
            if(b&16) *d++=*s++; else *d++;
        }
        else
        {
            for(i=0;i<5;i++)
                *d++=*s++;
        }
      }
    }
}

int main(int argc,char *argv[])
{
    FILE    *f,*d;
    unsigned char  xmid[18],name[21],tname[21],order[256],iname[23];
    unsigned short versio,songlen,restart,chn,ptns,insts,amiga,tempo,bpm,
                   samples;
    unsigned hsize,isize,shsize,ssize;
    static PTN     ptn[256];
    static SMP     smp[128];
    char    tmp;
    int     n,i,j;

    char    *id[]= {"M.K.","6CHN","8CHN","10CH","12CH","14CH","16CH",
                    "18CH","20CH","22CH","24CH","26CH","28CH","30CH","32CH"};
    int     chns[]={4,6,8,10,12,14,16,18,20,22,24,26,28,30,32,-1};

    int     ptable[]={0,
            1712,1616,1524,1440,1356,1280,1208,1140,1076,1016, 960, 906,
            856, 808, 762, 720, 678, 640, 604, 570, 538, 508, 480, 453,
            428, 404, 381, 360, 339, 320, 302, 285, 269, 254, 240, 226,
            214, 202, 190, 180, 170, 160, 151, 143, 135, 127, 120, 113,
            107, 101,  95,  90,  85,  80,  75,  71,  67,  63,  60,  56,
        };

    if(argc!=3)
    {
        fprintf(stderr,"musagi source.xm dest.s\n");
        return(1);
    }
    f=fopen(argv[1],"rb");
    if(f==NULL)
    {
        fprintf(stderr,"File not found\n");
        return(1);
    }

    for(n=0;n<31;n++)
    {
        smp[n].len=smp[n].vol=smp[n].loop=smp[n].looplen=0;
        smp[n].finetune=0;
        smp[n].name="";
    }

    for(n=0;n<17;n++)
        xmid[n]=U8();
    xmid[17]=0;
    if(strcmp(xmid,"Extended Module: "))
    {
        fprintf(stderr,"Not an XM\n",xmid);
        return(1);
    }
    puts("It is an XM indeed");

    for(n=0;n<20;n++)
    {
        tmp=U8();
        name[n]=(charset(tmp))?tmp:' ';
    }
    name[20]=0;
    printf("Name:%s\n",name);

    U8();
    for(n=0;n<20;n++)
    {
        tmp=U8();
        tname[n]=(charset(tmp))?tmp:' ';
    }
    tname[20]=0;
    printf("Tracker name:%s\n",tname);

    versio=U16();
    printf("Version:%x\n",versio);
    hsize=U32();
    printf("Header size:%d\n",hsize);
    songlen=U16();
    printf("Song length:%d\n",songlen);
    restart=U16();
    printf("Restart:%d\n",restart);
    chn=U16();
    printf("Channels:%d\n",chn);
    ptns=U16();
    printf("Patterns:%d\n",ptns);
    insts=U16();
    printf("Instruments:%d\n",insts);
    if(U16()&1)
        amiga=0;
    else
        amiga=1;
    printf("Linear periods:%s\n",(amiga)?"No. This won't work.":"Yes.");
    tempo=U16();
    printf("Tempo:%d\n",tempo);
    bpm=U16();
    printf("BPM:%d\n",bpm);

    for(n=0;n<256;n++)
        order[n]=U8();

    for(n=0;n<ptns;n++)
    {
        printf("Pattern %d:",n+1);
        U32();  // Skip some BS
        U8();
        ptn[n].rows=U16();
        ptn[n].packed=U16();
        if(ptn[n].packed==0)
            printf("empty pattern. ");
        ptn[n].pdata=malloc(ptn[n].packed);
        for(i=0;i<ptn[n].packed;i++)
            ptn[n].pdata[i]=U8();

        if(ptn[n].rows!=64)
            printf("Warning! Length not 64 rows, ");
        printf("%d rows\n",ptn[n].rows);

        convptn(&ptn[n],chn);
    }

    fclose(f);
    f=fopen(argv[2],"w");
    if(f==NULL)
    {
        printf("File error creating: %s\n",argv[2]);
        return(-1);
    }

    printf("Writing %s...\n",argv[2]);

    fprintf(f,"; Song repeat and order\n");
    fprintf(f,"repeat\tdw\torder+%d\n",restart*2);
    fprintf(f,"order");
    for(n=0;n<songlen;n++) // Song order
        fprintf(f,"\tdw\tpatt%d\n",order[n]);
    fprintf(f,"\tdw\t0\n");

    fprintf(f,"\n; Patterns\n");
    /* Write patterns */
    for(n=0;n<ptns && n<64;n++)
    {
        fprintf(f,"patt%d",n);
        unsigned char *s;
        unsigned char b1,b2,b3,b4;
        unsigned period,note;
        s=ptn[n].data;

        for(i=0;i<64;i++)
        {
            fprintf(f,"\tdb\t");
            if(i>=ptn[n].rows) // Add extra rows...
            {
                fprintf(f,"0,0,0, 0,0,0, 0,0,0");
            }
            else
            {
                for(j=0;j<3;j++,s+=5)
                {
                    fprintf(f,"%d,$%02x,$%02x",s[0],s[3],s[4]);
                    if(j!=2)
                        fprintf(f,", ");
                }
                s+=5; // Skip the 4th channel
            }
            fprintf(f,"\n");
        }
        fprintf(f,"\n");
    }

    fclose(f);

    return(0);
}
