; The JR-200 player part. Feel free to extend :)
;
; 26.3.2012: Full 8 octaves
; 21.3.2012: Added tremor (Txx)
; 18.3.2012: Initial version

	cpu	6800
	* = $1000

; Set a 50 Hz clock hook
	sei

	ldab	#13
	stab	$c817
	ldab	#86
	stab	$c818

	ldx	#hook
	stx	$10c

	cli
jumi	bra	jumi

; Timer interrupt hook
hook	ldaa	#1
	staa	$ca00

	dec	count
	beq	newrow
	jmp	norow

; A new row! Need to do several things
newrow	inc	row
	ldaa	row
	cmpa	#64
	bne	nopatt

; A new pattern! Gosh!
	clra
	staa	row
	ldx	pos
	inx
	inx
	stx	pos
	ldx	,x
	bne	loadptr

; Need to jump to the repeat point
	ldx	repeat
	stx	pos
	ldx	,x

loadptr	stx	rowptr
	bra	notes

nopatt	ldx	rowptr ; Next row
	inx
	inx
	inx
	inx
	inx
	inx
	inx
	inx
	inx
	stx	rowptr

; Handle new notes and once-per-row commands
; 1st channel
notes	ldx	rowptr
	ldaa	,x
	beq	nonote1
	staa	note1
	staa	base1

nonote1	ldaa	1,x
	beq	arp1
	cmpa	#$c
	beq	vol1
	cmpa	#$f
	beq	spd1
	cmpa	#$1d
	beq	trm1
	bra	chn2

trm1	ldaa	base1
	staa	note1
	bra	chn2

vol1	clra
	staa	note1
	bra	chn2

spd1	ldab	2,x
	stab	speed
	bra	chn2

arp1	ldab	2,x
	beq	chn2
	ldaa	base1
	staa	note1

; 2nd channel
chn2	ldaa	3,x
	beq	nonote2
	staa	note2
	staa	base2

nonote2	ldaa	4,x
	beq	arp2
	cmpa	#$c
	beq	vol2
	cmpa	#$f
	beq	spd2
	cmpa	#$1d
	beq	trm2
	bra	chn3

trm2	ldaa	base2
	staa	note2
	bra	chn3

vol2	clra
	staa	note2
	bra	chn3

spd2	ldab	5,x
	stab	speed
	bra	chn3

arp2	ldab	5,x
	beq	chn3
	ldaa	base2
	staa	note2

; 3rd channel
chn3	ldaa	6,x
	beq	nonote3
	staa	note3
	staa	base3

nonote3	ldaa	7,x
	beq	arp3
	cmpa	#$c
	beq	vol3
	cmpa	#$f
	beq	spd3
	cmpa	#$1d
	beq	trm3
	bra	rowdone

trm3	ldaa	base3
	staa	note3
	bra	rowdone

vol3	clra
	staa	note3
	bra	rowdone

spd3	ldab	8,x
	stab	speed
	bra	rowdone

arp3	ldab	8,x
	beq	rowdone
	ldaa	base3
	staa	note3

rowdone	ldaa	speed	; Reset tick counter
	staa	count

	clra		; Clear arpeggio and termor pos
	staa	arpc1
	staa	arpc2
	staa	arpc3

	staa	trmc1
	staa	trmc2
	staa	trmc3

	bra	getout

; No new row this time, just check the necessary commands
norow	ldx	rowptr

; 1st channel fx (just arpeggio and tremor for now)
fx1	ldaa	1,x
	cmpa	#$1d
	beq	trem1
	cmpa	#0
	bne	fx2
	ldaa	2,x
	beq	fx2
	ldx	#arpc1
	bsr	arppi
	bra	fx2

trem1	ldaa	2,x
	ldx	#trmc1
	bsr	tremmi

fx2	ldaa	4,x
	cmpa	#$1d
	beq	trem2
	cmpa	#0
	bne	fx3
	ldaa	5,x
	beq	fx3
	ldx	#arpc2
	bsr	arppi
	bra	fx3

trem2	ldaa	5,x
	ldx	#trmc2
	bsr	tremmi

fx3	ldaa	7,x
	cmpa	#$1d
	beq	trem3
	cmpa	#0
	bne	getout
	ldaa	8,x
	beq	getout
	ldx	#arpc3
	bsr	arppi
	bra	getout

trem3	ldaa	8,x
	ldx	#trmc3
	bsr	tremmi

getout	bsr	setfreq
	clra
	staa	$ca00
	rts

; Take care of the arpeggio notes (X=counter address, A=param)
arppi	inc	,x
	ldab	,x
	cmpb	#1
	beq	param1
	cmpb	#2
	beq	param2

	clrb ; Bare note, copy base to current
	stab	,x
	ldaa	3,x
	staa	6,x
	rts

param1	lsra ; High 4 bits = plus to base
	lsra
	lsra
	lsra
	adda	3,x
	staa	6,x
	rts

param2	anda	#15 ; Low bits = plus to base
	adda	3,x
	staa	6,x
	rts

; Take care of the tremor buzz (X=counter address, A=param)
tremmi	inc	,x
	tab
	lsrb
	lsrb
	lsrb
	lsrb
	cmpb	,x	; Note still on
	bpl	trdone

	anda	#15
	aba
	clrb		; Silence
	stab	9,x

	cmpa	,x	; Wrap the counter
	bpl	trout

	ldaa	#-1
	staa	,x
	rts

trdone	ldaa	6,x
	staa	9,x
trout	rts

; Routine to set the freqs of all channels
setfreq	ldaa	note1 ; Channel 1
	beq	silent1

	cmpa	prev1 ; Don't retrig
	beq	freq2
	staa	prev1

	asla
	staa	tmp+1
	ldx	tmp
	ldaa	,x
	staa	$c81a
	ldaa	1,x
	staa	$c81b
	ldaa	#$e
	staa	$c819
	bra	freq2

silent1	clra
	staa	$c819
	staa	prev1

freq2	ldaa	note2 ; Don't retrig, channel 2
	cmpa	prev2
	beq	freq3
	staa	prev2

	asla
	staa	tmp2+1
	ldx	tmp2
	ldaa	,x
	staa	$c813
	ldaa	1,x
	staa	$c812

freq3	ldaa	note3 ; Don't retrig, channel 3
	cmpa	prev3
	beq	frdone
	staa	prev3

	asla
	staa	tmp2+1
	ldx	tmp2
	ldaa	,x
	staa	$c815
	ldaa	1,x
	staa	$c814

frdone	rts

row	db	63
rowptr	dw	0
count	db	1
speed	db	6
pos	dw	order-2

tmp	dw	frequs ; For indexing the 256-aligned arrays
tmp2	dw	frequs2

trmc1	db	0 ; Tremor counters
trmc2	db	0
trmc3	db	0

arpc1	db	0 ; Arpeggio counters
arpc2	db	0
arpc3	db	0

base1	db	0 ; Base note for arpeggios
base2	db	0
base3	db	0

note1	db	0 ; Current note
note2	db	0
note3	db	0

prev1	db	0 ; Previous played note so that we don't retrig all the time
prev2	db	0
prev3	db	0

	include biisi.inc
	include frequs.inc
